import os
import shutil
import subprocess
import tempfile
import unittest

from click.testing import CliRunner

from pre_commit_vauxoo.cli import main


class TestPreCommitVauxoo(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.original_work_dir = os.getcwd()
        self.tmp_dir = tempfile.mkdtemp(suffix='_pre_commit_vauxoo')
        os.chdir(self.tmp_dir)
        self.runner = CliRunner()
        src_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'resources')
        self.create_dummy_repo(src_path, self.tmp_dir)
        self.maxDiff = None

    def create_dummy_repo(self, src_path, dest_path):
        subprocess.call(["git", "init", "--initial-branch=main", dest_path])
        dest_subpath = os.path.join(dest_path, os.path.basename(src_path))
        shutil.copytree(src_path, dest_subpath)
        subprocess.call(["git", "add", dest_subpath])

    def tearDown(self):
        super().tearDown()
        # change to original work dir
        if os.path.isdir(self.original_work_dir):
            os.chdir(self.original_work_dir)
        # Cleanup temporary files
        if os.path.isdir(self.tmp_dir) and self.tmp_dir != '/':
            shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_basic(self):
        os.environ['INCLUDE_LINT'] = 'resources'
        os.environ['PRECOMMIT_AUTOFIX'] = '1'
        result = self.runner.invoke(main, [])
        self.assertEqual(result.exit_code, 0, "Exited with error %s" % result)

    def test_chdir(self):
        self.runner = CliRunner()
        os.environ['PRECOMMIT_AUTOFIX'] = '1'
        os.chdir("resources")
        result = self.runner.invoke(main, [])
        self.assertEqual(result.exit_code, 0, "Exited with error %s" % result)

    def test_exclude_lint_path(self):
        self.runner = CliRunner()
        os.chdir("resources")
        os.environ['PRECOMMIT_AUTOFIX'] = '1'
        os.environ['EXCLUDE_LINT'] = 'resources/module_example1/models'
        result = self.runner.invoke(main, [])
        self.assertEqual(result.exit_code, 0, "Exited with error %s" % result)

    def test_disable_lints(self):
        self.runner = CliRunner()
        os.environ['DISABLE_PYLINT_CHECKS'] = 'import-error'
        result = self.runner.invoke(main, [])
        self.assertEqual(result.exit_code, 0, "Exited with error %s" % result)

    def test_exclude_autofix(self):
        self.runner = CliRunner()
        os.chdir("resources")
        os.environ['PRECOMMIT_AUTOFIX'] = '1'
        os.environ['EXCLUDE_AUTOFIX'] = 'resources/module_example1/demo/'
        result = self.runner.invoke(main, [])
        self.assertEqual(result.exit_code, 0, "Exited with error %s" % result)


if __name__ == '__main__':
    unittest.main()



{
    'name': "module_example1",
    'summary': """
        Short (1 phrase/line) summary of the module's purpose, used as
        subtitle on modules listing or apps.openerp.com""",
    'author': 'Vauxoo,Odoo Community Association (OCA)',
    'license': 'AGPL-3',
    'website': "http://www.yourcompany.com",
    # Categories can be used to filter modules in modules listing
    # Check https://github.com/odoo/odoo/blob/15.0/odoo/addons/base/data/ir_module_category_data.xml
    # for the full list
    'category': 'Uncategorized',
    'version': '15.0.1.0.0',
    # any module necessary for this one to work correctly
    'depends': ['base'],
    # always loaded
    'data': [
        # 'security/ir.model.access.csv',
        'views/views.xml',
    ],
    # only loaded in demonstration mode
    'demo': [
        'demo/demo.xml',
    ],
}



"""Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mpre_commit_vauxoo` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``pre_commit_vauxoo.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``pre_commit_vauxoo.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""


import contextlib
import os

import click

from pre_commit_vauxoo import pre_commit_vauxoo


def source_variables():
    # Overwrite os.environ with variables.sh file only if it was not already defined
    repo_dirname = pre_commit_vauxoo.get_repo()
    envdict = pre_commit_vauxoo.envfile2envdict(repo_dirname)
    for var, value in envdict.items():
        if var in os.environ:
            continue
        os.environ[var] = value


@contextlib.contextmanager
def env_clear():
    """Clear environment variables after finished the method"""
    old_environ = os.environ.copy()
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


def monkey_patch_make_context():
    """monkey patch to run source variables.sh before to parse the click.options"""
    original_make_context = click.core.BaseCommand.make_context

    def custom_make_context(*args, **kwargs):
        with env_clear():
            source_variables()
            return original_make_context(*args, **kwargs)

    click.core.BaseCommand.make_context = custom_make_context


def strcsv2tuple(strcsv, lower=False):
    if isinstance(strcsv, tuple):
        # Crazy!! but click==8.0.1 is sending "value" transformed
        strcsv = ','.join(strcsv)
    strcsv = strcsv and strcsv.strip() or ""
    if not strcsv:
        return ()
    items = ()
    for item in strcsv.split(','):
        item = item.strip()
        if lower:
            item = item.lower()
        items += (item,)
    return items


class CSVChoice(click.Choice):
    envvar_list_splitter = ','

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name += ' CSV'

    def convert(self, value, param, ctx):
        values = ()
        # case_sensitive parameter is not compatible with click 6.6
        # So we can do it manually here since that we are transforming it
        for v in strcsv2tuple(value, lower=True):
            values += (super().convert(v, param, ctx),)
        return values


class CSVPath(click.Path):
    envvar_list_splitter = ','

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name += ' CSV'

    def convert(self, value, param, ctx):
        values = ()
        for v in strcsv2tuple(value):
            try:
                new_value = super().convert(v, param, ctx)
            except click.exceptions.BadParameter:
                # The envvar are using path based on root repo path
                new_value = os.path.join(pre_commit_vauxoo.get_repo(), v)
                new_value = super().convert(new_value, param, ctx)
            values += (new_value,)
        return values


def merge_tuples(ctx, param, value):
    """Convert (('value1', 'value2'), ('value3')) to ('value1', 'value2', 'value3')
    It is useful for csv separated by comma parameters but using multiple number of args
    """
    if value is None:
        return value
    values = ()
    for v in value:
        if not isinstance(v, tuple):
            v = (v,)
        values += v
    return values


new_extra_kwargs = {}
try:
    if tuple(map(int, click.__version__.split('.'))) >= (7, 0):
        # It is only compatible for click >= 7.0 but it is not a big deal if it is not enabled
        # For record, dockerv image is using click 6.6 version
        new_extra_kwargs["show_envvar"] = True
except (TypeError, ValueError, AttributeError):  # pylint: disable=except-pass
    pass

monkey_patch_make_context()


@click.command()
# click 6.6 used in dockerv doesn't support to use envvar for click.argument :(
# More info https://github.com/pallets/click/issues/714 workaround using option instead.
@click.option(
    "--paths",
    "-p",
    multiple=True,
    envvar="INCLUDE_LINT",
    type=CSVPath(exists=True),
    callback=merge_tuples,
)
@click.option(
    "--overwrite",
    "-w",
    envvar="PRECOMMIT_OVERWRITE_CONFIG_FILES",
    default=True,
    show_default=True,
    help="Overwrite configuration files. "
    "If True, existing configuration files into the project will be overwritten. "
    "If False, then current files will be used, if they exist.",
    **new_extra_kwargs,
)
@click.option(
    "--exclude-autofix",
    "-x",
    type=CSVPath(exists=True),
    multiple=True,
    callback=merge_tuples,
    envvar="EXCLUDE_AUTOFIX",
    help="Exclude paths on which to run the autofix pre-commit configuration, separated by commas",
    **new_extra_kwargs,
)
@click.option(
    "--exclude-lint",
    "-l",
    type=CSVPath(exists=True),
    multiple=True,
    callback=merge_tuples,
    envvar="EXCLUDE_LINT",
    help="Paths to exclude checks, separated by commas.",
    **new_extra_kwargs,
)
@click.option(
    "--disable-pylint-checks",
    '-d',
    type=str,
    callback=merge_tuples,
    envvar="DISABLE_PYLINT_CHECKS",
    help="Pylint checks to disable, separated by commas.",
    **new_extra_kwargs,
)
@click.option(
    "--autofix",
    "-f",
    envvar="PRECOMMIT_AUTOFIX",
    is_flag=True,
    default=False,
    show_default=True,
    help="Run pre-commit with autofix configuration to change the source code. "
    "Overwrite -c option to '-c mandatory -c optional -c fix' ",
    **new_extra_kwargs,
)
@click.option(
    "--precommit-hooks-type",
    "-t",
    type=CSVChoice(["mandatory", "optional", "fix", "all"]),
    default=["mandatory", "optional"],
    multiple=True,
    callback=merge_tuples,
    show_default=True,
    envvar="PRECOMMIT_HOOKS_TYPE",
    help="Pre-commit configuration file to run hooks, separated by comma. "
    "*Mandatory: Stable hooks that needs to be fixed (Affecting build status). "
    "*Optional: Optional hooks that could be fixed later. (No affects build status). "
    "*Fix: Hooks auto fixing source code (Affects build status). "
    "*All: All configuration files to run hooks. ",
    **new_extra_kwargs,
)
def main(paths, overwrite, exclude_autofix, exclude_lint, disable_pylint_checks, autofix, precommit_hooks_type):
    """PATHS are the specific filenames to run hooks on separated by commas.
    Also, it can be defined using environment variable INCLUDE_LINT
    [default: Current directory]
    """
    pre_commit_vauxoo.main(
        paths, overwrite, exclude_autofix, exclude_lint, disable_pylint_checks, autofix, precommit_hooks_type
    )



import logging

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, _NOTHING, DEFAULT = range(10)
# The background is set with 40 plus the number of the color, and the foreground with 30
# These are the sequences needed to get colored output
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"
COLOR_PATTERN = "%s%s%%s%s" % (COLOR_SEQ, COLOR_SEQ, RESET_SEQ)
LEVEL_COLOR_MAPPING = {
    logging.DEBUG: (BLUE, DEFAULT),
    logging.INFO: (GREEN, DEFAULT),
    logging.WARNING: (YELLOW, DEFAULT),
    logging.ERROR: (RED, DEFAULT),
    logging.CRITICAL: (WHITE, RED),
}
FORMAT_STR = "%(asctime)s %(levelname)s %(name)s: %(message)s"


def colorized_msg(msg, level):
    fg_color, bg_color = LEVEL_COLOR_MAPPING.get(level, (GREEN, DEFAULT))
    colorized_msg_str = COLOR_PATTERN % (30 + fg_color, 40 + bg_color, msg)
    return colorized_msg_str


class ColoredFormatter(logging.Formatter):
    def format(self, record):
        level_colorized = colorized_msg(record.levelname, record.levelno)
        record.levelname = level_colorized
        return logging.Formatter.format(self, record)


logger = logging.getLogger("pre-commit-vauxoo")
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

ch.setFormatter(ColoredFormatter(FORMAT_STR))

logger.addHandler(ch)

if __name__ == '__main__':
    logger.info("testing....")



# from odoo import models, fields, api


# class module_example1(models.Model):
#     _name = 'module_example1.module_example1'
#     _description = 'module_example1.module_example1'

#     name = fields.Char()
#     value = fields.Integer()
#     value2 = fields.Float(compute="_value_pc", store=True)
#     description = fields.Text()
#
#     @api.depends('value')
#     def _value_pc(self):
#         for record in self:
#             record.value2 = float(record.value) / 100



import logging
import os
import re
import subprocess
import sys

from . import logging_colored

_logger = logging.getLogger("pre-commit-vauxoo")

re_export = re.compile(
    r"^(?P<export>export|EXPORT)( )+"
    + r"(?P<variable>[\w]*)[ ]*[\=][ ]*[\"\']{0,1}"
    + r"(?P<value>[\w\.\-\_/\$\{\}\:,\(\)\#\* ]*)[\"\']{0,1}",
    re.M,
)


def get_repo():
    repo_root = subprocess.check_output(["git", "rev-parse", "--show-toplevel"]).decode(sys.stdout.encoding).strip()
    repo_root = os.path.abspath(os.path.realpath(repo_root))
    return repo_root


def get_files(path):
    ls_files = subprocess.check_output(["git", "ls-files", "--", path]).decode(sys.stdout.encoding).strip()
    ls_files = ls_files.splitlines()
    return ls_files


def copy_cfg_files(
    precommit_config_dir, repo_dirname, overwrite, exclude_lint, disable_pylint_checks, exclude_autofix
):
    exclude_lint_regex = ""
    exclude_autofix_regex = ""
    if exclude_lint:
        exclude_lint_regex = "(%s)|" % "|".join(
            [re.escape(exclude_path.strip()) for exclude_path in exclude_lint if exclude_path and exclude_path.strip()]
        )
    if exclude_autofix:
        exclude_autofix_regex = "(%s)|" % "|".join(
            [
                re.escape(exclude_path.strip())
                for exclude_path in exclude_autofix
                if exclude_path and exclude_path.strip()
            ]
        )
    _logger.info("Copying configuration files 'cp -rnT %s/ %s/", precommit_config_dir, repo_dirname)
    for fname in os.listdir(precommit_config_dir):
        if not fname.startswith(".") and fname != "pyproject.toml":
            # all configuration files are hidden
            continue
        src = os.path.join(precommit_config_dir, fname)
        if not os.path.isfile(src):
            # if it is not a file skip
            continue
        dst = os.path.join(repo_dirname, fname)
        if not overwrite and os.path.isfile(dst):
            # Use the custom files defined in the repo
            _logger.warning("Using custom file %s", dst)
            continue
        with open(src, "r") as fsrc, open(dst, "w") as fdst:
            for line in fsrc:
                if fname.startswith(".pre-commit-config") and "# EXCLUDE_LINT" in line:
                    line = ""
                    if exclude_lint:
                        _logger.info("Applying EXCLUDE_LINT=%s to %s", exclude_lint, dst)
                        line += "    %s\n" % exclude_lint_regex
                    if fname == ".pre-commit-config-autofix.yaml" and exclude_autofix:
                        _logger.info("Applying EXCLUDE_AUTOFIX=%s to %s", exclude_autofix, dst)
                        line += "    %s\n" % exclude_autofix_regex
                if disable_pylint_checks and fname.startswith(".pre-commit-config") and "--disable=R0000" in line:
                    line = line.replace("R0000", ','.join(disable_pylint_checks))
                fdst.write(line)


def envfile2envdict(repo_dirname, source_file="variables.sh", no_overwrite_environ=True):
    """Simulate load the Vauxoo standard file 'source variables.sh' command in python
    return dictionary {environment_variable: value}
    """
    envdict = {}
    source_file_path = os.path.join(repo_dirname, source_file)
    if not os.path.isfile(source_file_path):
        _logger.info("Skipping 'source %s' file because it was not found", source_file_path)
        return envdict
    with open(source_file_path) as f_source_file:
        _logger.info("Running 'source %s'", source_file_path)
        for line in f_source_file:
            line_match = re_export.match(line)
            if not line_match:
                continue
            if no_overwrite_environ and line_match["variable"] in os.environ:
                continue
            envdict.update({line_match["variable"]: line_match["value"]})
    return envdict


def subprocess_call(command, *args, **kwargs):
    cmd_str = ' '.join(command)
    _logger.debug("Running command: %s", cmd_str)
    return subprocess.call(command, *args, **kwargs)


# There are a lot of if validations in this method. It is expected for now.
# pylint: disable=too-complex
def main(
    include_lint,
    overwrite,
    exclude_autofix,
    exclude_lint,
    disable_pylint_checks,
    autofix,
    precommit_hooks_type,
    do_exit=True,
):
    repo_dirname = get_repo()
    cwd = os.path.abspath(os.path.realpath(os.getcwd()))

    root_dir = os.path.abspath(os.path.dirname(__file__))

    precommit_config_dir = os.path.join(root_dir, "cfg")

    copy_cfg_files(
        precommit_config_dir,
        repo_dirname,
        overwrite,
        exclude_lint,
        disable_pylint_checks,
        exclude_autofix,
    )
    if autofix:
        precommit_hooks_type = ("mandatory", "optional", "fix")
    elif not precommit_hooks_type:
        precommit_hooks_type = ("mandatory", "optional")

    _logger.info("Installing pre-commit hooks")
    cmd = ["pre-commit", "install-hooks", "--color=always"]
    pre_commit_cfg_mandatory = os.path.join(repo_dirname, ".pre-commit-config.yaml")
    pre_commit_cfg_optional = os.path.join(repo_dirname, ".pre-commit-config-optional.yaml")
    pre_commit_cfg_autofix = os.path.join(repo_dirname, ".pre-commit-config-autofix.yaml")
    if "mandatory" in precommit_hooks_type or 'all' in precommit_hooks_type:
        subprocess_call(cmd + ["-c", pre_commit_cfg_mandatory])
    if "optional" in precommit_hooks_type or 'all' in precommit_hooks_type:
        subprocess_call(cmd + ["-c", pre_commit_cfg_optional])
    if "fix" in precommit_hooks_type or 'all' in precommit_hooks_type:
        subprocess_call(cmd + ["-c", pre_commit_cfg_autofix])

    status = 0
    cmd = ["pre-commit", "run", "--color=always"]
    if cwd != repo_dirname:
        cwd_short = os.path.relpath(cwd, repo_dirname)
        if include_lint:
            _logger.warning(
                "Ignored path configured '%s'. Use 'cd %s' and run the same command again to use configured path",
                ','.join(include_lint),
                repo_dirname,
            )
        _logger.warning("Running in current directory '%s'", cwd_short)
        files = get_files(cwd)
        if not files:
            raise UserWarning("Not files detected in current path %s" % cwd_short)
        cmd.extend(["--files"] + files)
    elif include_lint:
        _logger.info("Running only for INCLUDE_LINT=%s", include_lint)
        included_files = []
        for included_path in include_lint:
            included_files += get_files(included_path) or (included_path,)
        cmd.extend(["--files"] + included_files)
    else:
        cmd.append("--all")
    all_status = {}

    if "fix" in precommit_hooks_type or 'all' in precommit_hooks_type:
        _logger.info("%s AUTOFIX CHECKS %s", "-" * 25, "-" * 25)
        _logger.info("Running autofix checks (affect status build but you can autofix them locally)")
        autofix_status = subprocess_call(cmd + ["-c", pre_commit_cfg_autofix])
        status += autofix_status
        test_name = 'Autofix checks'
        all_status[test_name] = {'status': autofix_status}
        if autofix_status != 0:
            _logger.error("%s reformatted", test_name)
            all_status[test_name]['level'] = logging.ERROR
            all_status[test_name]['status_msg'] = "Reformatted"
        else:
            _logger.info("%s passed!", test_name)
            all_status[test_name]['level'] = logging.INFO
            all_status[test_name]['status_msg'] = "Passed"
        _logger.info("-" * 66)

    if "mandatory" in precommit_hooks_type or 'all' in precommit_hooks_type:
        _logger.info("%s MANDATORY CHECKS %s", "*" * 25, "*" * 25)
        _logger.info("Running mandatory checks (affect status build)")
        mandatory_status = subprocess_call(cmd + ["-c", pre_commit_cfg_mandatory])
        status += mandatory_status
        test_name = 'Mandatory checks'
        all_status[test_name] = {'status': mandatory_status}
        if status != 0:
            _logger.error("%s failed", test_name)
            all_status[test_name]['level'] = logging.ERROR
            all_status[test_name]['status_msg'] = "Failed"
        else:
            _logger.info("%s passed!", test_name)
            all_status[test_name]['level'] = logging.INFO
            all_status[test_name]['status_msg'] = "Passed"

    if "optional" in precommit_hooks_type or 'all' in precommit_hooks_type:
        _logger.info("*" * 68)
        _logger.info("%s OPTIONAL CHECKS %s", "~" * 25, "~" * 25)
        _logger.info("Running optional checks (does not affect status build)")
        status_optional = subprocess_call(cmd + ["-c", os.path.join(repo_dirname, pre_commit_cfg_optional)])
        test_name = 'Optional checks'
        all_status[test_name] = {'status': status_optional}
        if status_optional != 0:
            _logger.warning("Optional checks failed")
            all_status[test_name]['level'] = logging.WARNING
            all_status[test_name]['status_msg'] = "Failed"
        else:
            _logger.info("Optional checks passed!")
            all_status[test_name]['level'] = logging.INFO
            all_status[test_name]['status_msg'] = "Passed"
        _logger.info("~" * 67)

    print_summary(all_status)
    if do_exit:
        sys.exit(0 if status == 0 else 1)


def print_summary(all_status):
    summary_msg = ["+" + "=" * 39]
    summary_msg.append("|  Tests summary:")
    summary_msg.append("|" + "-" * 39)
    for test_name, test_result in all_status.items():
        outcome = (
            logging_colored.colorized_msg(test_result['status_msg'], test_result['level'])
            if test_result['status'] != 0
            else logging_colored.colorized_msg(test_result['status_msg'], test_result['level'])
        )
        summary_msg.append("| {:<28}{}".format(test_name, outcome))
    summary_msg.append("+" + "=" * 39)
    _logger.info("Tests summary\n%s", '\n'.join(summary_msg))


if __name__ == "__main__":
    main()
